###############################################################################
## $Id: c172p-dual-control.nas,v 1.36 2009-05-24 16:51:31 anders Exp $
##
## Nasal for dual control of the c172p over the multiplayer network.
##
##  Copyright (C) 2007 - 2008  Anders Gidenstam  (anders(at)gidenstam.org)
##  This file is licensed under the GPL license version 2 or later.
##
###############################################################################

# Renaming (almost :)
var DCT = dual_control_tools;

######################################################################
# Pilot/copilot aircraft identifiers. Used by dual_control.
var pilot_type   = "Aircraft/c172p-dual-control/Models/c172p-dual-command.xml";
var copilot_type = "Aircraft/c172p-dual-control/Models/c172p-copilot.xml";

######################################################################
# MP enabled properties.
# NOTE: These must exist very early during startup - put them
#       in the -set.xml file.
var pilot_rudder_mpp   = "surface-positions/rudder-pos-norm";
var pilot_elevator_mpp = "surface-positions/elevator-pos-norm";
var pilot_aileron_mpp  = "surface-positions/left-aileron-pos-norm";
var pilot_flap_mpp     = "surface-positions/flap-pos-norm";
var pilot_elevator_trim_mpp = "rotors/main/rpm";
var pilot_throttle_mpp = "rotors/main/blade[0]/position-deg";
var pilot_mixture_mpp  = "rotors/main/blade[0]/flap-deg";
var pilot_switches_mpp = "sim/model/variant";

var pilot_hsi_head_mpp          = "rotors/tail/blade[0]/position-deg";
var pilot_tc_turn_rate_mpp      = "rotors/main/blade[3]/position-deg";
var pilot_tc_slip_skid_mpp      = "rotors/main/blade[3]/flap-deg";
var pilot_ai_pitch_mpp          = "sim/hitches/aerotow/tow/end-force-x";
var pilot_ai_roll_mpp           = "sim/hitches/aerotow/tow/end-force-y";
var pilot_ai_hoffset_mpp        = "sim/hitches/aerotow/tow/end-force-z";

var pilot_TDM_mpp  = "sim/multiplay/generic/string[0]";

var copilot_rudder_mpp   = "surface-positions/rudder-pos-norm";
var copilot_elevator_mpp = "surface-positions/elevator-pos-norm";
var copilot_aileron_mpp  = "surface-positions/left-aileron-pos-norm";
var copilot_flap_mpp     = "surface-positions/flap-pos-norm";
var copilot_elevator_trim_mpp = "rotors/main/rpm";
var copilot_lbrake_mpp   = "surface-positions/right-aileron-pos-norm";
var copilot_rbrake_mpp   = "surface-positions/speedbrake-pos-norm";
var copilot_throttle_mpp = "rotors/main/blade[0]/position-deg";
var copilot_mixture_mpp  = "rotors/main/blade[0]/flap-deg";
var copilot_switches_mpp = "sim/model/variant";
var copilot_switches2_mpp = "sim/multiplay/generic/int[0]";

var copilot_TDM_mpp = pilot_TDM_mpp;

######################################################################
# Useful instrument related property paths.

# Flight controls
var rudder_cmd   = "controls/flight/rudder";
var elevator_cmd = "controls/flight/elevator";
var aileron_cmd  = "controls/flight/aileron";
var throttle_cmd = "controls/engines/engine/throttle";
var mixture_cmd  = "controls/engines/engine/mixture";
var elevator_trim_cmd = "controls/flight/elevator-trim";

# Instruments
var hsi_heading = "instrumentation/heading-indicator/indicated-heading-deg";
var altimeter_setting = "instrumentation/altimeter/setting-inhg";
var tc_turn_rate      = "instrumentation/turn-indicator/indicated-turn-rate";
var tc_slip_skid      = "instrumentation/slip-skid-ball/indicated-slip-skid";
var rpm               = "engines/engine/rpm";
var ai_pitch   = "instrumentation/attitude-indicator/indicated-pitch-deg";
var ai_roll    = "instrumentation/attitude-indicator/indicated-roll-deg";
var ai_hoffset = "instrumentation/attitude-indicator/horizon-offset-deg";

# Switches
var carb_heat         = "controls/anti-ice/engine/carb-heat";
var pitot_heat        = "controls/anti-ice/pitot-heat";
var taxi_light        = "controls/lighting/taxi-light";
var landing_lights    = "controls/lighting/landing-lights";
var nav_lights        = "controls/lighting/nav-lights";
var beacon_light      = "controls/lighting/beacon";
var strobe_light      = "controls/lighting/strobe";

######################################################################
# Slow state properties for replication.

###############################################################################
# Pilot MP property mappings and specific copilot connect/disconnect actions.

var l_dual_control         = "fdm/jsbsim/fcs/dual-control";

var l_throttle_pos         = "fdm/jsbsim/fcs/throttle-pos-norm";
var l_pilot_throttle_cmd   = "fdm/jsbsim/fcs/throttle-cmd-norm";
var l_copilot_throttle_cmd = "fdm/jsbsim/fcs/copilot-throttle-cmd-norm";
var l_shared_throttle_cmd  = "fdm/jsbsim/fcs/shared-throttle-cmd-norm";

var l_mixture_pos          = "fdm/jsbsim/fcs/mixture-pos-norm";
var l_pilot_mixture_cmd    = "fdm/jsbsim/fcs/mixture-cmd-norm";
var l_copilot_mixture_cmd  = "fdm/jsbsim/fcs/copilot-mixture-cmd-norm";
var l_shared_mixture_cmd   = "fdm/jsbsim/fcs/shared-mixture-cmd-norm";

var l_elevator_trim_cmd    = "controls/flight/elevator-trim";

######################################################################
# Used by dual_control to set up the mappings for the pilot.
var pilot_connect_copilot = func (copilot) {
  # Make sure dual-control is activated in the FDM FCS.
  setprop(l_dual_control, 1);

  # KX-165 NavComm. NavComm 1 is owned by pilot, 2 by copilot.
  KX165.make_master(0);
  KX165.make_slave_to(1, copilot);

  # KI-206 VOR indicator. Nav 0 is owned by pilot. Nav 1 by copilot.
  KI206.make_master(0);
  KI206.make_slave_to(1, copilot);

  # KR-87 ADF radio
  KR87.make_master(0);

  return [
    ##################################################
    # Copilot main flight control
    DCT.Translator.new
      (copilot.getNode("surface-positions/elevator-pos-norm"),
       props.globals.getNode("/fdm/jsbsim/fcs/copilot-elevator-cmd-norm")),
    DCT.Translator.new
      (copilot.getNode("surface-positions/left-aileron-pos-norm"),
       props.globals.getNode("/fdm/jsbsim/fcs/copilot-aileron-cmd-norm")),
    DCT.Translator.new
      (copilot.getNode("surface-positions/rudder-pos-norm"),
       props.globals.getNode("/fdm/jsbsim/fcs/copilot-rudder-cmd-norm")),
    ##################################################
    # Copilot elevator trim control
    DCT.DeltaAdder.new
      (copilot.getNode(copilot_elevator_trim_mpp),
       props.globals.getNode(l_elevator_trim_cmd)),
    ##################################################
    # Copilot flap control
    DCT.EdgeTrigger.new
      (copilot.getNode(copilot_flap_mpp),
       func (v) { controls.flapsDown(v); },
       func (v) { controls.flapsDown(-v); }),
    ##################################################
    # Copilot engine control inputs
    DCT.Translator.new
      (copilot.getNode(copilot_throttle_mpp),
       props.globals.getNode(l_copilot_throttle_cmd)),
    DCT.Translator.new
      (copilot.getNode(copilot_mixture_mpp),
       props.globals.getNode(l_copilot_mixture_cmd)),
    # Engine sharing
    DCT.MostRecentSelector.new
      (props.globals.getNode(l_pilot_throttle_cmd),
       props.globals.getNode(l_copilot_throttle_cmd),
       props.globals.getNode(l_shared_throttle_cmd),
       threshold = 0.02),
    DCT.MostRecentSelector.new
      (props.globals.getNode(l_pilot_mixture_cmd),
       props.globals.getNode(l_copilot_mixture_cmd),
       props.globals.getNode(l_shared_mixture_cmd),
       threshold = 0.02),
    ##################################################
    # Copilot brake control
    DCT.EdgeTrigger.new
      (copilot.getNode(copilot_lbrake_mpp),
       func (v) { controls.applyBrakes(v, -1); },
       func (v) { controls.applyParkingBrake(v); }),
    DCT.EdgeTrigger.new
      (copilot.getNode(copilot_rbrake_mpp),
       func (v) { controls.applyBrakes(v, 1); },
       func (v) { }),
    ##################################################
    # Decode copilot cockpit switch states.
    #   NOTE: Actions are only triggered on change.
    DCT.SwitchDecoder.new
      (copilot.getNode(copilot_switches_mpp),
       [func (b) {
         props.globals.getNode(carb_heat).setValue(b?1:0);
        },
        func (b) {
         props.globals.getNode(taxi_light).setValue(b?1:0);
        },
        func (b) {
         props.globals.getNode(pitot_heat).setValue(b?1:0);
        },
        func (b) {
         props.globals.getNode(nav_lights).setValue(b?1:0);
        },
        func (b) {
         props.globals.getNode(landing_lights).setValue(b?1:0);
        },
        func (b) {
         props.globals.getNode(beacon_light).setValue(b?1:0);
        },
        func (b) {
         props.globals.getNode(strobe_light).setValue(b?1:0);
        }
       ] ~
       #  7
       ########################################
       #  KX-165 NavComm 1
       KX165.master_receive_slave_buttons(0) ~
       #  17
       #  KI-206 Nav 1
       KI206.master_receive_slave_buttons(0) ~
       #  19
       #  KR-87 ADF Radio
       KR87.master_receive_slave_buttons(0)
       #  24
      ),
    DCT.SwitchDecoder.new
      (copilot.getNode(copilot_switches2_mpp),
       # KAP140
       kap140.master_receive_slave_buttons()
       # 11
      ),
    ##################################################
    # Set up TDM reception of slow state properties.
    DCT.TDMDecoder.new
      (copilot.getNode(copilot_TDM_mpp),
       #  Copilot NavComm 2
       KX165.slave_receive_master_state(1) ~
       KI206.slave_receive_master_state(1)
      ),
    ##################################################
    # Control settings
    DCT.Translator.new
      (props.globals.getNode(l_throttle_pos),
       props.globals.getNode(pilot_throttle_mpp)),
    DCT.Translator.new
      (props.globals.getNode(l_mixture_pos),
       props.globals.getNode(pilot_mixture_mpp)),
    DCT.Translator.new
      (props.globals.getNode(l_elevator_trim_cmd),
       props.globals.getNode(pilot_elevator_trim_mpp)),
    ##################################################
    # HSI indicated heading.
    DCT.Translator.new
      (props.globals.getNode(hsi_heading),
       props.globals.getNode(pilot_hsi_head_mpp)),
    ################################################## 
    # Turn coordinator turn rate.
    DCT.Translator.new
      (props.globals.getNode(tc_turn_rate),
       props.globals.getNode(pilot_tc_turn_rate_mpp)),
    ##################################################
    # Turn coordinator slip skid.
    DCT.Translator.new
      (props.globals.getNode(tc_slip_skid),
       props.globals.getNode(pilot_tc_slip_skid_mpp)),
    ##################################################
    # Attitude Indicator.
    DCT.Translator.new
      (props.globals.getNode(ai_pitch),
       props.globals.getNode(pilot_ai_pitch_mpp)),
    DCT.Translator.new
      (props.globals.getNode(ai_roll),
       props.globals.getNode(pilot_ai_roll_mpp)),
     # Not sure why the next property fails.
#    DCT.Translator.new
#      (props.globals.getNode(ai_hoffset),
#       props.globals.getNode(pilot_ai_hoffset_mpp)),
    ##################################################
    # Encoding of on/off switches.
    DCT.SwitchEncoder.new
      ([props.globals.getNode("controls/anti-ice/engine/carb-heat"),
        props.globals.getNode("controls/lighting/taxi-light"),
        props.globals.getNode("controls/anti-ice/pitot-heat"),
        props.globals.getNode("controls/lighting/nav-lights"),
        props.globals.getNode("controls/lighting/landing-lights"),
        props.globals.getNode("controls/lighting/beacon"),
        props.globals.getNode("controls/lighting/strobe")
       ] ~
       #  7 NavComm 2 buttons.
       KX165.slave_send_buttons(1) ~
       #  17
       KI206.slave_send_buttons(1)
       #  19
       ,
       props.globals.getNode(pilot_switches_mpp)),
    ##################################################
    # Set up TDM transmission of slow state properties.
    DCT.TDMEncoder.new
      ([props.globals.getNode(altimeter_setting)
       ] ~
       #  NavComm 1 state
       KX165.master_send_state(0) ~
       #  VOR indicator 1 state
       KI206.master_send_state(0) ~
       #  ADF 1 state
       KR87.master_send_state(0)
       ,
       props.globals.getNode(pilot_TDM_mpp),
      )
  ];
}

######################################################################
var pilot_disconnect_copilot = func {
  # Reset copilot controls. Slightly dangerous.
  setprop("/fdm/jsbsim/fcs/copilot-elevator-cmd-norm", 0.0);
  setprop("/fdm/jsbsim/fcs/copilot-rudder-cmd-norm", 0.0);
  setprop("/fdm/jsbsim/fcs/copilot-aileron-cmd-norm", 0.0);
  setprop(l_dual_control, 0);

  # KX-165 NavComm. Regain control of NavComm 1.
  KX165.make_master(1);
  # KI-206 VOR indicator. Regain control of Nav 1.
  KI206.make_master(1);
}


###############################################################################
# Copilot MP property mappings and specific pilot connect/disconnect actions.

######################################################################
# Used by dual_control to set up the mappings for the copilot.
var copilot_connect_pilot = func (pilot) {
  # Initialize Nasal wrappers for copilot pick anaimations.
  set_copilot_wrappers(pilot);

  return [
    ##################################################
    # Map properties for animation and sound.
    # Map combined throttle cmd for animation.
    DCT.Translator.new
      (pilot.getNode(pilot_throttle_mpp),
       pilot.getNode("fdm/jsbsim/fcs/throttle-pos-norm")),
    # Map combined mixture cmd for animation.
    DCT.Translator.new
      (pilot.getNode(pilot_mixture_mpp),
       pilot.getNode("fdm/jsbsim/fcs/mixture-pos-norm")),
    # Map elevator trim for animation
    DCT.Translator.new
      (pilot.getNode(pilot_elevator_trim_mpp),
       pilot.getNode(elevator_trim_cmd)),
    # Map /engines/engine/rpm for sound and RPM instrument.
    DCT.Translator.new
      (pilot.getNode(rpm),
       props.globals.getNode(rpm)),
    # Map flap-pos-norm for flap sound.
    DCT.Translator.new
      (pilot.getNode(pilot_flap_mpp),
       props.globals.getNode("/fdm/jsbsim/fcs/flap-pos-norm", 1)),
    ##################################################
    # Map airspeed for airspeed indicator. This is cheating!
    DCT.Translator.new
      (pilot.getNode("velocities/true-airspeed-kt"),
       props.globals.getNode("/instrumentation/" ~
                             "airspeed-indicator/indicated-speed-kt", 1)),
    ##################################################
    # Map HSI indicated heading for animation.
    DCT.Translator.new
      (pilot.getNode(pilot_hsi_head_mpp),
       props.globals.getNode(hsi_heading)),
    ##################################################
    # Map turn coordinator turn rate for animation.
    DCT.Translator.new
      (pilot.getNode(pilot_tc_turn_rate_mpp),
       props.globals.getNode(tc_turn_rate)),
    # Map turn coordinator slip skid.
    DCT.Translator.new
      (pilot.getNode(pilot_tc_slip_skid_mpp),
       props.globals.getNode(tc_slip_skid)),
    ##################################################
    # Map Attitude Indicator for animation.
    DCT.Translator.new
      (pilot.getNode(pilot_ai_pitch_mpp),
       props.globals.getNode(ai_pitch)),
    DCT.Translator.new
      (pilot.getNode(pilot_ai_roll_mpp),
       props.globals.getNode(ai_roll)),
    DCT.Translator.new
      (pilot.getNode(pilot_ai_hoffset_mpp),
       props.globals.getNode(ai_hoffset)),
    ##################################################
    # Map M877 clock properties to pilot 3d model. Local replica.
#    DCT.Translator.new
#      (props.globals.getNode("instrumentation/clock/m877/mode"),
#       pilot.getNode("instrumentation/clock/m877/mode")),
#    DCT.Translator.new
#      (props.globals.getNode("instrumentation/clock/m877/indicated-hour"),
#       pilot.getNode("instrumentation/clock/m877/indicated-hour")),
#    DCT.Translator.new
#      (props.globals.getNode("instrumentation/clock/m877/indicated-min"),
#       pilot.getNode("instrumentation/clock/m877/indicated-min")),
    ##################################################
    # Decode pilot cockpit switch states for animation and control.
    DCT.SwitchDecoder.new
      (pilot.getNode(pilot_switches_mpp),
       [func (b) {
         pilot.getNode(carb_heat).setValue(b?1:0);
         props.globals.getNode(carb_heat).setValue(b?1:0);
        },
        func (b) {
         pilot.getNode(taxi_light).setValue(b?1:0);
         props.globals.getNode(taxi_light).setValue(b?1:0);
        },
        func (b) {
         pilot.getNode(pitot_heat).setValue(b?1:0);
         props.globals.getNode(pitot_heat).setValue(b?1:0);
        },
        func (b) {
         pilot.getNode(nav_lights).setValue(b?1:0);
         props.globals.getNode(nav_lights).setValue(b?1:0);
        },
        func (b) {
         pilot.getNode(landing_lights).setValue(b?1:0);
         props.globals.getNode(landing_lights).setValue(b?1:0);
        },
        func (b) {
         pilot.getNode(beacon_light).setValue(b?1:0);
         props.globals.getNode(beacon_light).setValue(b?1:0);
        },
        func (b) {
         pilot.getNode(strobe_light).setValue(b?1:0);
         props.globals.getNode(strobe_light).setValue(b?1:0);
        }
       ] ~
       #  7
       ########################################
       #  KX-165 NavComm 2
       KX165.master_receive_slave_buttons(1) ~
       #  17
       #  KI-206 Nav 2
       KI206.master_receive_slave_buttons(1)
       # 19
      ),
    ##################################################
    # Set up TDM reception of slow state properties.
    DCT.TDMDecoder.new
      (pilot.getNode(pilot_TDM_mpp),
       [func (v) {
          props.globals.getNode(altimeter_setting).setValue(v);
        }
       ] ~
       KX165.slave_receive_master_state(0) ~
       KI206.slave_receive_master_state(0) ~
       KR87.slave_receive_master_state(0)       
      ),
    ##################################################
    # Map copilot flight controls to MP properties.
    DCT.Translator.new
      (props.globals.getNode(rudder_cmd),
       props.globals.getNode(copilot_rudder_mpp), factor = -1),
    DCT.Translator.new
      (props.globals.getNode(elevator_cmd),
       props.globals.getNode(copilot_elevator_mpp)),
    DCT.Translator.new
      (props.globals.getNode(aileron_cmd),
       props.globals.getNode(copilot_aileron_mpp)),
    DCT.Translator.new
      (props.globals.getNode(throttle_cmd),
       props.globals.getNode(copilot_throttle_mpp)),
    DCT.Translator.new
      (props.globals.getNode(mixture_cmd),
       props.globals.getNode(copilot_mixture_mpp)),
    DCT.Translator.new
      (props.globals.getNode(elevator_trim_cmd),
       props.globals.getNode(copilot_elevator_trim_mpp)),
    ##################################################
    # Encoding of on/off switches.
    DCT.SwitchEncoder.new
      ([props.globals.getNode(carb_heat),
        props.globals.getNode(taxi_light),
        props.globals.getNode(pitot_heat),
        props.globals.getNode(nav_lights),
        props.globals.getNode(landing_lights),
        props.globals.getNode(beacon_light),
        props.globals.getNode(strobe_light)
       ] ~
       #  7 NavComm 1
       KX165.slave_send_buttons(0) ~
       #  17 VOR 1
       KI206.slave_send_buttons(0) ~
       #  19 ADF 1
       KR87.slave_send_buttons(0)
       #  24
       ,
       props.globals.getNode(copilot_switches_mpp)),
    DCT.SwitchEncoder.new
      (kap140.slave_send_buttons()
       # 11
       ,
       props.globals.getNode(copilot_switches2_mpp)),
    ##################################################
    # Set up TDM transmission of slow state properties.
    DCT.TDMEncoder.new
      (KX165.master_send_state(1) ~
       KI206.master_send_state(1),
       props.globals.getNode(copilot_TDM_mpp),
      )
  ];
}

var copilot_disconnect_pilot = func {
}

######################################################################
# Copilot Nasal wrappers

var set_copilot_wrappers = func (pilot) {
  #######################################################
  # controls.nas wrapper overrides.

  # Flap control input is -1 for step decrease; 1 for step increase; 0 idle
  controls.flapsDown = func {
    var val = arg[0];
    if(val > 1) { val = 1 } elsif(val < -1) { val = -1 }
    setprop(copilot_flap_mpp, val);
  }

  # Brake control
  controls.applyBrakes = func(v, which = 0) {
    if (which <= 0) { setprop(copilot_lbrake_mpp, v); }
    if (which >= 0) { setprop(copilot_rbrake_mpp, v); }
  }

  # FIXME: Odd.
  controls.applyParkingBrake = func(v) {
    if (!v) { return; }
    setprop(copilot_lbrake_mpp, -v);
    return 0;
  }

  #######################################################
  # Pick animation action overrides.
  kap140.make_slave_to(pilot);
  kap140.animate_aimodel(pilot);

  # KX-165 NavComm. NavComm 1 is owned by pilot, 2 by copilot.
  KX165.make_slave_to(0, pilot);
  KX165.make_master(1);
  KX165.animate_aimodel(0, pilot);
  KX165.animate_aimodel(1, pilot);

  # KI-206 VOR indicator. Nav 1 is owned by pilot. Nav 2 by copilot.
  KI206.make_slave_to(0, pilot);
  KI206.make_master(1);
  KI206.animate_aimodel(0, pilot);
  KI206.animate_aimodel(1, pilot);

  # KDI-572 DME indicator
  KDI572.animate_aimodel(0, pilot);

  # KR-87 ADF radio. Owned by pilot.
  KR87.make_slave_to(0, pilot);
  KR87.animate_aimodel(0, pilot);
}

